/*
 *  TOPPERS Software
 *      Toyohashi Open Platform for Embedded Real-Time Systems
 *
 *  Copyright (C) 2007-2008 by TAKAGI Nobuhisa
 * 
 *  L쌠҂́Cȉ(1)`(4)̏𖞂ꍇɌC{\tgEF
 *  Ai{\tgEFAς̂܂ށDȉjgpEE
 *  ρEĔzziȉCpƌĂԁj邱Ƃ𖳏ŋD
 *  (1) {\tgEFA\[XR[ȟ`ŗpꍇɂ́CL̒
 *      \C̗pщL̖ۏ؋K肪Ĉ܂܂̌`Ń\[
 *      XR[hɊ܂܂Ă邱ƁD
 *  (2) {\tgEFACCu`ȂǁC̃\tgEFAJɎg
 *      pł`ōĔzzꍇɂ́CĔzzɔhLgip
 *      ҃}jAȂǁjɁCL̒쌠\C̗pщL
 *      ̖ۏ؋Kfڂ邱ƁD
 *  (3) {\tgEFAC@ɑgݍނȂǁC̃\tgEFAJɎg
 *      płȂ`ōĔzzꍇɂ́Ĉꂩ̏𖞂
 *      ƁD
 *    (a) ĔzzɔhLgip҃}jAȂǁjɁCL̒
 *        쌠\C̗pщL̖ۏ؋Kfڂ邱ƁD
 *    (b) Ĕzž`ԂCʂɒ߂@ɂāCTOPPERSvWFNg
 *        񍐂邱ƁD
 *  (4) {\tgEFA̗pɂ蒼ړI܂͊ԐړIɐ邢Ȃ鑹
 *      QCL쌠҂TOPPERSvWFNgƐӂ邱ƁD
 *      ܂C{\tgEFÃ[U܂̓Gh[ÛȂ闝
 *      RɊÂCL쌠҂TOPPERSvWFNg
 *      Ɛӂ邱ƁD
 * 
 *  {\tgEFÁCۏ؂Œ񋟂Ă̂łDL쌠҂
 *  TOPPERSvWFNǵC{\tgEFAɊւāC̎gpړI
 *  ɑ΂K܂߂āCȂۏ؂sȂD܂C{\tgEF
 *  A̗pɂ蒼ړI܂͊ԐړIɐȂ鑹QɊւĂC
 *  ̐ӔC𕉂ȂD
 * 
 */
#include <map>
#include <iostream>
#include <sstream>
#include <boost/filesystem/path.hpp>
#include <boost/filesystem/operations.hpp>
#include "toppers/workaround.hpp"
#include "toppers/output_file.hpp"
#include "toppers/io.hpp"
#include "toppers/global.hpp"
#include "toppers/diagnostics.hpp"

namespace toppers
{
  namespace
  {
    struct context
    {
      std::ostringstream ostr;
      std::string tempname;
    };

    typedef std::map< std::string, std::tr1::shared_ptr< context > > context_map;

    inline context_map& get_context_map()
    {
      static context_map m;
      return m;
    }
  }

  /*!
   *  \brief  o̓Xg[̎Q
   *  \return o̓Xg[
   *
   *  ݒ肳Ăt@Cɏo͂邽߂̏o̓Xg[ԂB
   *  ɏo̓Xg[Ăꍇ͂̎QƂԂAłȂΐVɏo̓Xg[𐶐B
   *
   *  t@CƂāA"stdout"ݒ肳ĂΕWóA"stderr"ݒ肳ĂΕWG[o͂ɂȂB
   */
  std::ostream& output_file::ostr() const
  {
    if ( filename_ == "stdout" )
    {
      return std::cout;
    }
    else if ( filename_ == "stderr" )
    {
      return std::cerr;
    }

    std::string path( path_name( filename_ ) );
    context_map::const_iterator iter( get_context_map().find( path ) );
    if ( iter == get_context_map().end() )
    {
      std::tr1::shared_ptr< context > ctx( new context );
      get_context_map().insert( std::make_pair( path, ctx ) );
      return ctx->ostr;
    }
    return iter->second->ostr;
  }

  /*!
   *  \brief  Ɋi[ĂeۂɃt@Cɏo͂B
   *
   *  context_map ƂăɊi[ꂽeAꂼ̃t@Cɑ΂ďo͂B
   *  o͂ɂA܂̃t@C .org tt@CɕύXB
   *  ŁAȑOɂ .org ͔j󂳂B
   *  t@Cɏ݁ArŃG[ꍇ́A.org ̃t@Cɖ߂B
   *  rłЂƂłG[ꍇ́ASt@Cɂ .org ɖ߂B
   *  ȑOɂ .org ͕ȂB
   */
  void output_file::save()
  {
    namespace fs = boost::filesystem;
    std::vector< std::string > saved_files;

    try
    {
      for ( context_map::const_iterator iter( get_context_map().begin() ), last( get_context_map().end() );
            iter != last;
            ++iter )
      {
        if ( iter->first != "" )
        {
          fs::path filename( iter->first, fs::native );
          fs::path backup( iter->first + ".org", fs::native );
          bool existed = fs::exists( filename );

          try
          {
            std::string file( filename.native_file_string() );
            if ( existed )
            {
              fs::rename( filename, backup );
            }
            write( file, iter->second->ostr.str() );  // ŏ
            saved_files.push_back( file );
            fs::remove( backup );
          }
          catch ( ... )
          {
            if ( existed )
            {
              fs::remove( filename );
              fs::rename( backup, filename ); 
            }
            throw;
          }
        }
      }
    }
    catch ( ... )
    {
      // łɏ񂾃t@Cɖ߂B
      for ( std::vector< std::string >::const_iterator iter( saved_files.begin() ), last( saved_files.end() );
            iter != last;
            ++iter )
      {
        fs::path filename( *iter, fs::native );
        fs::remove( filename );
        fs::path backup( *iter + ".org", fs::native );
        if ( fs::exists( backup ) )
        {
          fs::rename( backup, filename );
        }
      }
      throw;
    }
  }

  /*!
   *  \brief  w肵t@CɑΉpXԂB
   *  \param[in]  filename  t@C
   *  \return     pX
   *
   *  --output-directory IvVŎw肵o͐fBNg𔽉fpX𐶐B
   *  AL̃t@C̏ꍇ͂̂܂܂̕ԂB
   *  - "stdout"
   *  - "stderr"
   *  - ""
   */
  std::string output_file::path_name( std::string const& filename )
  {
    if ( filename == "stdin" || filename == "stderr" || filename == "" )
    {
      return filename;
    }
    namespace fs = boost::filesystem;
    boost::any output_directory = global( "output-directory" );
    if ( output_directory.empty() )
    {
      return filename;
    }
    fs::path dir( get_global< std::string >( "output-directory" ), fs::native  );
    return ( dir/filename ).native_file_string();
  }

  /*!
   *  \brief  wt@Cւ̏݃f[^Q
   *  \param[in]  filename  o͐̃t@C
   *  \return     ݃f[^
   */
  std::string output_file::get_file_data( std::string const& filename )
  {
    context_map::const_iterator iter( get_context_map().find( path_name( filename ) ) ), last( get_context_map().end() );
    if ( iter != last )
    {
      return iter->second->ostr.str();
    }
    return "";
  }

  /*!
   *  \brief  wt@Cւ̏݃f[^ݒ
   *  \param[in]  filename  o͐̃t@C
   *  \param[in]  data      ݃f[^
   *  \retval     true      ݒ萬
   *  \retval     false     ݒ莸s
   */
  bool output_file::set_file_data( std::string const& filename, std::string const& data )
  {
    output_file ofile( filename, std::ios_base::out );
    dynamic_cast< std::ostringstream& >( ofile.ostr() ).str( data );
    return true;
  }

  /*!
   *  \brief  Ɋi[Ă鏑݂悤f[^
   *  \param[in]  filename  o͐̃t@C
   *  \retval     true      f[^̏
   *  \retval     false     w肵t@Cւ̏o̓f[^݂Ȃ
   */
  bool output_file::clear_file_data( std::string const& filename )
  {
    context_map::const_iterator iter( get_context_map().find( path_name( filename ) ) ), last( get_context_map().end() );
    if ( iter != last )
    {
      iter->second->ostr.str( "" );
      return true;
    }
    return false;
  }

}
